from __future__ import annotations

import asyncio
import contextlib
import importlib
import inspect
import os
import pkgutil
import time
from argparse import ArgumentParser
from collections import namedtuple
from multiprocessing import Manager
from multiprocessing.managers import DictProxy
from typing import Any, Dict, List, Optional, Type, Union, cast

from pydantic import BaseModel

from .checkpointer import CheckpointContents, Checkpointer, NoCheckpointAvailableError
from .checkpointer.checkpointer import _checkpointer_classes
from .checkpointer.file import FileCheckpointer
from .eventbus import EventBus
from .eventbus.execution import ExecutionState, ExecutionStateMap, ExecutionStateTransition, SerializableExecutionEvent
from .eventbus.log import SerializableLogEvent
from .exceptions import CheckpointInvalidError, ExtensionsDirectoryNotFoundError, NotAPackageError, NoTasksLoadedError
from .executor import Executor
from .extensions.extension import Extension, _extension_classes
from .jobdefinition import (
    ConfigurationDefinition,
    ExtensionDefinition,
    JobDefinition,
    LoadParams,
    LoggerDefinition,
    TaskDefinition,
    _job_definition_classes,
)
from .loggers.logger import Logger, _logger_classes
from .task import Task

__all__ = ['Flowmancer']

ExecutorDetails = namedtuple('ExecutorDetails', 'instance dependencies')


# Need to explicitly manage loop in case multiple instances of Flowmancer are run.
@contextlib.contextmanager
def _create_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.stop()
    loop.close()


def _load_extensions_path(path: str, package_chain: Optional[List[str]] = None):
    if not path.startswith('/'):
        path = os.path.abspath(
            os.path.join(
                os.path.dirname(
                    os.path.abspath(inspect.stack()[-1][1])
                ),
                path
            )
        )

    if not os.path.exists(path):
        raise ExtensionsDirectoryNotFoundError(f"No such directory: '{path}'")
    if os.path.isfile(path):
        raise NotAPackageError(f"Only packages (directories) are allowed. The following is not a dir: '{path}'")
    if not os.path.exists(os.path.join(path, '__init__.py')):
        print(f"WARNING: The '{path}' dir is not a package (no __init__.py file found). Modules will not be imported.")

    if not package_chain:
        package_chain = [os.path.basename(path)]

    for x in pkgutil.iter_modules(path=[path]):
        try:
            print(f"importing: {'.'.join(package_chain+[x.name])}")
            importlib.import_module('.'.join(package_chain+[x.name]))
        except Exception as e:
            print(
                f"WARNING: Skipping import for '{'.'.join(package_chain+[x.name])}' due to {type(e).__name__}: {str(e)}"
            )
        if x.ispkg:
            _load_extensions_path(os.path.join(path, x.name), package_chain+[x.name])


class Flowmancer:
    def __init__(self, *, test: bool = False, debug: bool = False) -> None:
        manager = Manager()
        self._config: ConfigurationDefinition = ConfigurationDefinition()
        self._test = test
        self._debug = debug
        self._log_event_bus = EventBus[SerializableLogEvent](manager.Queue())
        self._execution_event_bus = EventBus[SerializableExecutionEvent]()
        self._shared_dict: DictProxy[str, Any] = manager.dict()
        self._executors: Dict[str, ExecutorDetails] = dict()
        self._states = ExecutionStateMap()
        self._registered_extensions: Dict[str, Extension] = dict()
        self._registered_loggers: Dict[str, Logger] = dict()
        self._checkpointer_instance: Checkpointer = FileCheckpointer()
        self._checkpointer_interval_seconds = 10.0
        self._extensions_interval_seconds = 0.25
        self._loggers_interval_seconds = 0.25
        self._synchro_interval_seconds = 0.25
        self._is_restart = False
        self._jobdef_vars: Dict[str, str] = dict()

    def set_jobdef_var(self, key: str, value: str) -> None:
        if not isinstance(key, str):
            raise TypeError(f'str expected for `key`, not {type(key)}')
        if not isinstance(value, str):
            raise TypeError(f'str expected for `value`, not {type(value)}')
        self._jobdef_vars[key] = value

    def unset_jobdef_var(self, key: str) -> None:
        if not isinstance(key, str):
            raise TypeError(f'str expected for `key`, not {type(key)}')
        del self._jobdef_vars[key]

    def start(self, default_jobdef_path: Optional[str] = None, default_jobdef_type: str = 'yaml') -> int:
        orig_cwd = os.getcwd()
        try:
            # Ensure any components, such as file loggers, work with respect to the .py file in which the `start`
            # command is invoked, which is usually the project root dir.
            app_root_dir = os.path.dirname(os.path.abspath(inspect.stack()[-1][1]))
            os.chdir(app_root_dir)
            if not self._test:
                self._process_cmd_args(orig_cwd, app_root_dir, default_jobdef_path, default_jobdef_type)
            if not self._executors:
                raise NoTasksLoadedError(
                    'No Tasks have been loaded! Please check that you have provided a valid Job Definition file.'
                )
            ret = asyncio.run(self._initiate())
            return ret
        finally:
            os.chdir(orig_cwd)

    async def _initiate(self) -> int:
        with _create_loop():
            root_event = asyncio.Event()
            if self._config.max_concurrency > 0:
                semaphore = asyncio.Semaphore(self._config.max_concurrency)
                for i in self._executors.values():
                    i.instance.semaphore = semaphore
            observer_tasks = self._init_extensions(root_event)
            executor_tasks = self._init_executors(root_event)
            logger_tasks = self._init_loggers(root_event)
            checkpoint_task = self._init_checkpointer(root_event)
            await asyncio.gather(*observer_tasks, *executor_tasks, *logger_tasks, checkpoint_task)
        return len(self._states[ExecutionState.FAILED]) + len(self._states[ExecutionState.DEFAULTED])

    def _validate_checkpoint(self, checkpoint: CheckpointContents) -> None:
        cp = set().union(*list(checkpoint.states.values()))
        ex = set(self._executors.keys())
        missing_in_checkpoint = ex.difference(cp)
        missing_in_executors = cp.difference(ex)
        if cp != ex:
            msg = 'Task names in Checkpoint do not match registered task names!'
            if missing_in_checkpoint:
                msg += '\nTasks MISSING in Checkpoint:'
                for n in missing_in_checkpoint:
                    msg += f'\n  * {n}'
            if missing_in_executors:
                msg += '\nUNKNOWN tasks in Checkpoint:'
                for n in missing_in_executors:
                    msg += 'f\n  * {n}'
            raise CheckpointInvalidError('Task names in Checkpoint do not match registered task names:')

    def _process_cmd_args(
        self,
        caller_cwd: str,
        app_root_dir: str,
        default_jobdef_path: Optional[str] = None,
        default_jobdef_type: str = 'yaml'
    ) -> None:
        # If a relative path is given in `start()`, it should be relative to the `start()` call. This is in contrast
        # to path given from command line, in which that case paths should be relative to where the command is executed.
        if default_jobdef_path and not default_jobdef_path.startswith('/'):
            default_jobdef_path = os.path.join(app_root_dir, default_jobdef_path)
        parser = ArgumentParser(description='Flowmancer job execution options.')
        parser.add_argument('-j', '--jobdef', action='store', dest='jobdef', default=default_jobdef_path)
        parser.add_argument('-t', '--type', action='store', dest='jobdef_type', default=default_jobdef_type)
        parser.add_argument('-r', '--restart', action='store_true', dest='restart', default=False)
        parser.add_argument('-d', '--debug', action='store_true', dest='debug', default=False)
        parser.add_argument('--skip', action='append', dest='skip', default=[])
        parser.add_argument('--run-to', action='store', dest='run_to')
        parser.add_argument('--run-from', action='store', dest='run_from')
        parser.add_argument('--max-concurrency', action='store', type=int, dest='max_concurrency')
        parser.add_argument('--var', action='append', dest='jobdef_vars', default=[])

        args = parser.parse_args()
        self._debug = args.debug

        for v in args.jobdef_vars:
            parts = v.split('=')
            if len(parts) <= 1:
                raise ValueError('`var` arguments must follow the pattern: <key>=<value>')
            self.set_jobdef_var(parts[0], '='.join(parts[1:]))

        if args.jobdef:
            jobdef_path = args.jobdef if args.jobdef.startswith('/') else os.path.join(caller_cwd, args.jobdef)
            self.load_job_definition(jobdef_path, app_root_dir, args.jobdef_type)

        if args.restart:
            try:
                cp = asyncio.run(self._checkpointer_instance.read_checkpoint(self._config.name))
                self._validate_checkpoint(cp)
                self._shared_dict.update(cp.shared_dict)
                esm = ExecutionStateMap.from_simple_dict(cp.states)
                for name in esm[ExecutionState.FAILED]:
                    self._executors[name].instance.is_restart = True
                completed = esm[ExecutionState.COMPLETED].copy()
                esm[ExecutionState.INIT].update(esm[ExecutionState.FAILED])
                esm[ExecutionState.INIT].update(esm[ExecutionState.ABORTED])
                esm[ExecutionState.INIT].update(esm[ExecutionState.RUNNING])
                esm[ExecutionState.INIT].update(esm[ExecutionState.PENDING])
                esm[ExecutionState.INIT].update(esm[ExecutionState.DEFAULTED])
                esm[ExecutionState.INIT].update(esm[ExecutionState.COMPLETED])
                esm[ExecutionState.FAILED].clear()
                esm[ExecutionState.ABORTED].clear()
                esm[ExecutionState.RUNNING].clear()
                esm[ExecutionState.PENDING].clear()
                esm[ExecutionState.DEFAULTED].clear()
                esm[ExecutionState.COMPLETED].clear()
                self._states = esm
                self._is_restart = True
                # Even though completed and don't require to be executed again, these tasks still need to trigger
                # state change from INIT -> COMPLETED for components watching the Execution Event Bus.
                for n in completed:
                    self._executors[n].instance.state = ExecutionState.COMPLETED
            except NoCheckpointAvailableError:
                self._is_restart = False
                print(f"No checkpoint file found for '{self._config.name}'. Starting new job.")

        # These override settings from JobsDefinition, if also defined there.
        if args.max_concurrency is not None:
            self._config.max_concurrency = args.max_concurrency

    def _is_failed(self) -> bool:
        return bool(
            self._states[ExecutionState.FAILED]
            or self._states[ExecutionState.DEFAULTED]
            or self._states[ExecutionState.ABORTED]
        )

    # ASYNC INITIALIZATIONS
    def _init_checkpointer(self, root_event) -> asyncio.Task:
        async def _write_checkpoint() -> None:
            await self._checkpointer_instance.write_checkpoint(
                self._config.name,
                CheckpointContents(
                    name=self._config.name,
                    states=self._states.to_simple_dict(),
                    shared_dict=self._shared_dict.copy()
                )
            )

        async def _pusher() -> None:
            last_write = 0
            await self._checkpointer_instance.on_create()
            if self._is_restart:
                await self._checkpointer_instance.on_restart()
            while True:
                if root_event.is_set():
                    break
                if (time.time() - last_write) >= self._checkpointer_interval_seconds:
                    await _write_checkpoint()
                await asyncio.sleep(self._synchro_interval_seconds)

            if self._is_failed():
                await _write_checkpoint()
                await self._checkpointer_instance.on_failure()
            else:
                await self._checkpointer_instance.on_success()
                await self._checkpointer_instance.clear_checkpoint(self._config.name)
            await self._checkpointer_instance.on_destroy()

        return asyncio.create_task(_pusher())

    def _init_executors(self, root_event: asyncio.Event) -> List[asyncio.Task]:
        async def _synchro() -> None:
            while not root_event.is_set():
                if (
                    not self._states[ExecutionState.INIT]
                    and not self._states[ExecutionState.PENDING]
                    and not self._states[ExecutionState.RUNNING]
                ):
                    root_event.set()
                else:
                    await asyncio.sleep(self._synchro_interval_seconds)

        return [asyncio.create_task(_synchro())] + [
            asyncio.create_task(dtl.instance.start())
            for name, dtl in self._executors.items()
            if name in self._states[ExecutionState.INIT]
        ]

    def _init_loggers(self, root_event: asyncio.Event) -> List[asyncio.Task]:
        if self._test:
            self._registered_loggers = dict()

        async def _write_logs() -> None:
            while not self._log_event_bus.empty():
                m = self._log_event_bus.get()
                for log in self._registered_loggers.values():
                    await log.update(m)

        async def _pusher() -> None:
            for log in self._registered_loggers.values():
                await log.on_create()
                if self._is_restart:
                    await log.on_restart()

            last_trigger = 0
            while True:
                if root_event.is_set():
                    break
                if (time.time() - last_trigger) >= self._loggers_interval_seconds:
                    await _write_logs()
                await asyncio.sleep(self._synchro_interval_seconds)

            await _write_logs()
            for log in self._registered_loggers.values():
                if self._is_failed():
                    await log.on_failure()
                else:
                    await log.on_success()
                await log.on_destroy()

        return [asyncio.create_task(_pusher())]

    def _init_extensions(self, root_event: asyncio.Event) -> List[asyncio.Task]:
        if self._test:
            self._registered_extensions = dict()

        async def _emit() -> None:
            while not self._execution_event_bus.empty():
                e = self._execution_event_bus.get()
                if self._debug:
                    print(e)
                if isinstance(e, ExecutionStateTransition):
                    self._states[e.to_state].add(e.name)
                    self._states[e.from_state].remove(e.name)
                for obs in self._registered_extensions.values():
                    await obs.update(e)

        async def _pusher() -> None:
            for obs in self._registered_extensions.values():
                await obs.on_create()
                if self._is_restart:
                    await obs.on_restart()

            last_trigger = 0
            while True:
                if root_event.is_set():
                    break
                if (time.time() - last_trigger) >= self._extensions_interval_seconds:
                    await _emit()
                await asyncio.sleep(self._synchro_interval_seconds)

            await _emit()
            for obs in self._registered_extensions.values():
                if self._is_failed():
                    await obs.on_failure()
                else:
                    await obs.on_success()
                await obs.on_destroy()

        return [asyncio.create_task(_pusher())]

    # EXECUTORS
    def _dependencies_are_valid(self) -> bool:
        for name, dtl in self._executors.items():
            for dep in dtl.dependencies:
                # Missing dependency
                if dep not in self._executors.keys():
                    return False
                # Self referencing dependency
                if dep == name:
                    return False
        return True

    def add_executor(
        self,
        name: str,
        task_class: Union[str, Type[Task]],
        deps: Optional[List[str]] = None,
        max_attempts: int = 1,
        backoff: int = 0,
        parameters: Dict[str, Any] = dict()
    ) -> None:
        async def await_dependencies() -> bool:
            for dep_name in self._executors[name].dependencies:
                d = self._executors[dep_name].instance
                await d.wait()
                if d.state in (ExecutionState.FAILED, ExecutionState.DEFAULTED, ExecutionState.ABORTED):
                    return False
            return True

        e = Executor(
            name=name,
            task_class=task_class,
            log_event_bus=self._log_event_bus,
            execution_event_bus=self._execution_event_bus,
            shared_dict=self._shared_dict,
            await_dependencies=await_dependencies,
            max_attempts=max_attempts,
            backoff=backoff,
            parameters=parameters
        )

        self._executors[name] = ExecutorDetails(instance=e, dependencies=(deps or []))
        self._states[ExecutionState.INIT].add(name)

    def load_job_definition(
        self,
        j: Union[JobDefinition, str],
        app_root_dir: str,
        jobdef_type: str = 'yaml'
    ) -> Flowmancer:
        if isinstance(j, JobDefinition):
            jobdef = j
        else:
            jobdef = _job_definition_classes[jobdef_type]().load(
                j, LoadParams(APP_ROOT_DIR=app_root_dir), self._jobdef_vars
            )

        # Configurations
        self._config = jobdef.config
        self._synchro_interval_seconds = jobdef.config.synchro_interval_seconds
        self._loggers_interval_seconds = jobdef.config.loggers_interval_seconds
        self._extensions_interval_seconds = jobdef.config.extensions_interval_seconds
        self._checkpointer_interval_seconds = jobdef.config.checkpointer_interval_seconds

        # Recursively import any modules found in the following paths in order to trigger the registration of any
        # decorated classes.
        for p in ['./tasks', './extensions', './loggers']:
            try:
                _load_extensions_path(p)
            except ExtensionsDirectoryNotFoundError:
                # Don't error on the absence of dirs that are searched by default.
                pass

        # Allow for missing dir exceptions for passed-in paths.
        for p in jobdef.config.extension_directories:
            _load_extensions_path(p)

        for p in jobdef.config.extension_packages:
            importlib.import_module(p)

        # Tasks
        for n, t in jobdef.tasks.items():
            self.add_executor(
                name=n,
                task_class=t.task,
                deps=t.dependencies,
                max_attempts=t.max_attempts,
                backoff=t.backoff,
                parameters=t.parameters
            )

        # Checkpointer
        self._checkpointer_instance = _checkpointer_classes[jobdef.checkpointer.checkpointer](
            **jobdef.checkpointer.parameters
        )

        # Observers
        for n, e in jobdef.extensions.items():
            self._registered_extensions[n] = _extension_classes[e.extension](**e.parameters)

        # Loggers
        for n, l in jobdef.loggers.items():
            self._registered_loggers[n] = _logger_classes[l.logger](**l.parameters)

        return self

    def get_job_definition(self) -> JobDefinition:
        j = JobDefinition(tasks=dict())
        j.config = self._config

        for n, e in self._executors.items():
            j.tasks[n] = TaskDefinition(
                task=type(e.instance.get_task_instance()).__name__,
                dependencies=[],
                parameters=e.instance.get_task_instance().dict()
            )

        for n, e in self._registered_extensions.items():
            j.extensions[n] = ExtensionDefinition(
                extension=type(e).__name__,
                parameters=cast(BaseModel, e).dict()
            )

        for n, l in self._registered_loggers.items():
            j.loggers[n] = LoggerDefinition(
                logger=type(l).__name__,
                parameters=cast(BaseModel, l).dict()
            )

        return j



import asyncio
import time
from collections import defaultdict
from typing import Dict

from rich.progress import Progress, TaskID

from ..executor import ExecutionState, ExecutionStateTransition, SerializableExecutionEvent
from .extension import Extension, extension


@extension
class RichProgressBar(Extension):
    class RichProgressBarState:
        def __init__(self) -> None:
            self.state_counts: Dict[ExecutionState, int]
            self.start_time: float
            self.progress: Progress
            self.event: asyncio.Event
            self.task: TaskID
            self.update_task: asyncio.Task

    _state: RichProgressBarState = RichProgressBarState()

    def _update_pbar(self, advance: int = 0) -> None:
        pending = self._state.state_counts[ExecutionState.PENDING]
        running = self._state.state_counts[ExecutionState.RUNNING]
        failed = (
            self._state.state_counts[ExecutionState.FAILED] + self._state.state_counts[ExecutionState.DEFAULTED]
        )
        completed = self._state.state_counts[ExecutionState.COMPLETED]
        total = (pending + running + failed + completed) or 100
        elapsed = int(time.time() - self._state.start_time)
        m = (
            f'Pending: {pending} - '
            + f'Running: {running} - '
            + f'Completed: {completed} - '
            + f'Failed: {failed} '
            + f'(Elapsed: {elapsed} sec.)'
        )
        self._state.progress.update(self._state.task, description=m, advance=advance, total=total)

    async def _continuous_update_pbar(self) -> None:
        while True:
            self._update_pbar()
            if self._state.event.is_set():
                # Need to imitate do-while to ensure pbar is updated one final time
                # before exiting loop due to Event set.
                break
            await asyncio.sleep(0.5)

    async def on_create(self) -> None:
        self._state.state_counts = defaultdict(lambda: 0)
        self._state.start_time = time.time()
        self._state.progress = Progress()
        self._state.event = asyncio.Event()
        self._state.task = self._state.progress.add_task('Pending: 0 - Running: 0 - Completed: 0 - Failed: 0')
        self._state.progress.start()

        loop = asyncio.get_event_loop()
        self._state.update_task = loop.create_task(self._continuous_update_pbar())

    async def on_destroy(self) -> None:
        self._state.event.set()
        await asyncio.gather(self._state.update_task)
        self._state.progress.stop()

    async def update(self, e: SerializableExecutionEvent) -> None:
        if isinstance(e, ExecutionStateTransition):
            from_state = ExecutionState(e.from_state)
            to_state = ExecutionState(e.to_state)
            self._state.state_counts[from_state] -= 1
            self._state.state_counts[to_state] += 1
            if to_state in (ExecutionState.FAILED, ExecutionState.COMPLETED, ExecutionState.DEFAULTED):
                self._update_pbar(1)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Set

from pydantic import BaseModel, ConfigDict

from ..lifecycle import AsyncLifecycle

_checkpointer_classes = dict()


def checkpointer(t: type[Checkpointer]) -> Any:
    if not issubclass(t, Checkpointer):
        raise TypeError(f'Must extend `Checkpointer` type: {t.__name__}')
    _checkpointer_classes[t.__name__] = t
    return t


class NoCheckpointAvailableError(Exception):
    pass


@dataclass
class CheckpointContents:
    name: str
    states: Dict[str, Set[str]]
    shared_dict: Dict[Any, Any]


class Checkpointer(ABC, BaseModel, AsyncLifecycle):
    model_config = ConfigDict(extra='forbid')

    @abstractmethod
    async def write_checkpoint(self, name: str, content: CheckpointContents) -> None:
        pass

    @abstractmethod
    async def read_checkpoint(self, name: str) -> CheckpointContents:
        pass

    @abstractmethod
    async def clear_checkpoint(self, name: str) -> None:
        pass



import glob
import os
import time
from datetime import datetime
from typing import Dict, TextIO

from ..eventbus.log import LogEndEvent, LogStartEvent, LogWriteEvent, SerializableLogEvent
from .logger import Logger, logger


class LogFileIsAlreadyOpen(Exception):
    pass


class LogFileNotOpen(Exception):
    pass


@logger
class FileLogger(Logger):
    class FileLoggerState:
        def __init__(self) -> None:
            self.file_handles: Dict[str, TextIO] = dict()
            self.ts_str = datetime.now().strftime('%Y-%m-%d.%H.%M.%S')

    _state: FileLoggerState = FileLoggerState()
    base_log_dir: str = './logs'
    retention_days: int = 10

    @property
    def log_dir(self) -> str:
        return f'{self.base_log_dir}/{self._state.ts_str}'

    async def on_create(self) -> None:
        os.makedirs(self.log_dir, exist_ok=True)

    async def update(self, msg: SerializableLogEvent) -> None:
        if isinstance(msg, LogStartEvent):
            f = self._state.file_handles.get(msg.name)
            if f and not f.closed:
                raise LogFileIsAlreadyOpen(f'Log file is already open for {msg.name}')
            self._state.file_handles[msg.name] = open(f'{self.log_dir}/{msg.name}.log', 'a')
        elif isinstance(msg, LogEndEvent):
            f = self._state.file_handles.get(msg.name)
            if f and not f.closed:
                f.close()
        elif isinstance(msg, LogWriteEvent):
            f = self._state.file_handles.get(msg.name)
            if not f or f.closed:
                raise LogFileNotOpen(f'Log file is not open for {msg.name}')
            template = '[{ts}] {sev} - {m}\n'
            f.write(template.format(
                sev=msg.severity,
                ts=msg.timestamp,
                m=msg.message
            ))

    async def on_destroy(self) -> None:
        for f in self._state.file_handles.values():
            if not f.closed:
                f.close()

        if self.retention_days < 0:
            return

        def _should_delete_file(fpath):
            return os.stat(fpath).st_mtime < (time.time() - (self.retention_days * 86400.0))

        def _should_delete_dir(dpath):
            return os.path.isdir(dpath) and not os.listdir(dpath)

        files = filter(_should_delete_file, glob.glob(f'{self.base_log_dir}/**/*.log'))
        # Type ignore hints due to mypy bugging out and detecting incorrect type...
        for f in files:  # type: ignore
            print(f'Deleting Log File: {f}')
            os.remove(f)  # type: ignore

        dirs = filter(_should_delete_dir, [f'{self.base_log_dir}/{p}' for p in os.listdir(self.base_log_dir)])
        for d in dirs:
            print(f'Deleting Directory: {d}')
            os.rmdir(d)



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections import defaultdict
from queue import Queue
from typing import Any, Dict, Generic, Optional, Type, TypeVar

from pydantic import BaseModel, ConfigDict

_event_classes: Dict[str, Dict[str, Type[SerializableEvent]]] = defaultdict(dict)


class NotASerializableEventError(Exception):
    pass


class NotADeserializableEventError(Exception):
    pass


class SerializableEvent(BaseModel, ABC):
    model_config = ConfigDict(extra='forbid')

    @classmethod
    @abstractmethod
    def event_group(cls) -> str:
        raise AttributeError('The `event_group` method has not been defined.')

    def serialize(self) -> str:
        try:
            return json.dumps({'group': self.event_group(), 'event': type(self).__name__, 'body': self.model_dump()})
        except Exception as e:
            raise NotASerializableEventError(str(e))

    @classmethod
    def deserialize(cls, e: str) -> Any:
        try:
            data = json.loads(e)
        except Exception as err:
            raise NotADeserializableEventError(str(err))
        if 'group' not in data or 'event' not in data or 'body' not in data:
            return UnknownEvent(content=e)
        else:
            return _event_classes[data['group']][data['event']](**data['body'])


class UnknownEvent(SerializableEvent):
    content: str

    @classmethod
    def event_group(cls) -> str:
        return 'UnknownEvent'


T = TypeVar('T', bound=SerializableEvent)


class EventBus(Generic[T]):
    __slots__ = ('_queue',)

    def __init__(self, q: Optional[Queue[str]] = None) -> None:
        self._queue = q or Queue()

    def put(self, m: T) -> None:
        self._queue.put(m.serialize())

    def get(self) -> T:
        return SerializableEvent.deserialize(self._queue.get())

    def empty(self) -> bool:
        return self._queue.empty()


def serializable_event(t: type[T]) -> type[T]:
    if not issubclass(t, SerializableEvent):
        raise TypeError(f'Must extend `SerializableEvent` type: {t.__name__}')
    _event_classes[t.event_group()][t.__name__] = t
    return t



class Lifecycle:
    def on_create(self) -> None:
        # Optional lifecycle method
        pass

    def on_restart(self) -> None:
        # Optional lifecycle method
        pass

    def on_success(self) -> None:
        # Optional lifecycle method
        pass

    def on_failure(self) -> None:
        # Optional lifecycle method
        pass

    def on_destroy(self) -> None:
        # Optional lifecycle method
        pass

    def on_abort(self) -> None:
        # Optional lifecycle method
        pass


class AsyncLifecycle:
    async def on_create(self) -> None:
        # Optional lifecycle method
        pass

    async def on_restart(self) -> None:
        # Optional lifecycle method
        pass

    async def on_success(self) -> None:
        # Optional lifecycle method
        pass

    async def on_failure(self) -> None:
        # Optional lifecycle method
        pass

    async def on_destroy(self) -> None:
        # Optional lifecycle method
        pass

    async def on_abort(self) -> None:
        # Optional lifecycle method
        pass



import requests

from ..extension import extension
from .notification import Notification


@extension
class PushoverNotification(Notification):
    app_token: str
    user_key: str

    async def send_notification(self, title: str, msg: str) -> None:
        headers = {'Content-type': 'application/x-www-form-urlencoded'}
        data = {'token': self.app_token, 'user': self.user_key, 'title': title, 'message': msg}
        requests.post('https://api.pushover.net/1/messages.json', headers=headers, data=data)



import pickle
import sqlite3
from contextlib import closing
from typing import Optional
from uuid import UUID, uuid4

from .checkpointer import CheckpointContents, Checkpointer, NoCheckpointAvailableError, checkpointer


@checkpointer
class SQLiteCheckpointer(Checkpointer):
    class SQLiteCheckpointerState:
        def __init__(self):
            self.con: Optional[sqlite3.Connection] = None
            self.uuid: UUID = uuid4()

    _state: SQLiteCheckpointerState = SQLiteCheckpointerState()
    checkpoint_database: str = './.flowmancer/checkpoint.db'

    @property
    def _con(self) -> sqlite3.Connection:
        if not self._state.con:
            self._state.con = sqlite3.connect(self.checkpoint_database)
            self._state.con.row_factory = sqlite3.Row
        return self._state.con

    async def on_create(self) -> None:
        self._con.execute('''
                    CREATE TABLE IF NOT EXISTS checkpoint (
                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                        uuid TEXT NOT NULL UNIQUE,
                        name TEXT NOT NULL,
                        start_ts DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP,
                        end_ts DATETIME,
                        checkpoint_contents BLOB
                    )
                    ''')

    async def on_destroy(self) -> None:
        if self._state.con:
            self._state.con.close()

    async def write_checkpoint(self, name: str, content: CheckpointContents) -> None:
        dumped_content = sqlite3.Binary(pickle.dumps(content))
        params = [str(self._state.uuid), name, dumped_content, dumped_content]
        self._con.execute('''
                          INSERT INTO checkpoint (uuid, name, checkpoint_contents) VALUES (?, ?, ?)
                          ON CONFLICT(uuid) DO UPDATE SET checkpoint_contents = ?
                          ''', params)
        self._con.commit()

    async def read_checkpoint(self, name: str) -> CheckpointContents:
        with closing(self._con.cursor()) as cur:
            # Check if the checkpoint table exists. In the event of first ever execution for the given DB, no such
            # table will exist and therefore no checkpoints exist.
            cur.execute("SELECT 1 FROM sqlite_master WHERE type = 'table' AND name = 'checkpoint'")
            row = cur.fetchone()
            if not row:
                raise NoCheckpointAvailableError(f'Checkpoint entry does not exist for: {name}')

            cur.execute('SELECT * FROM checkpoint WHERE name = ? ORDER BY id DESC LIMIT 1', [name])
            row = cur.fetchone()
            if row and not row['end_ts']:
                self._state.uuid = UUID(row['uuid'])
            else:
                raise NoCheckpointAvailableError(f'Checkpoint entry does not exist for: {name}')

            return pickle.loads(row['checkpoint_contents'])

    async def clear_checkpoint(self, _: str) -> None:
        self._con.execute(
            'UPDATE checkpoint SET end_ts = CURRENT_TIMESTAMP WHERE uuid = ?',
            [str(self._state.uuid)]
        )
        self._con.commit()



# noqa: F401
# Ensure implementations are registered
from . import database, file
from .checkpointer import CheckpointContents, Checkpointer, NoCheckpointAvailableError, checkpointer

__all__ = ['Checkpointer', 'CheckpointContents', 'NoCheckpointAvailableError', 'checkpointer']



from __future__ import annotations

from abc import ABC, abstractmethod

from pydantic import BaseModel, ConfigDict

from ..executor import SerializableExecutionEvent
from ..lifecycle import AsyncLifecycle

_extension_classes = dict()


def extension(t: type[Extension]):
    if not issubclass(t, Extension):
        raise TypeError(f'Must extend `Extension` type: {t.__name__}')
    _extension_classes[t.__name__] = t
    return t


class Extension(ABC, AsyncLifecycle, BaseModel):
    model_config = ConfigDict(extra='forbid')

    @abstractmethod
    async def update(self, e: SerializableExecutionEvent) -> None:
        pass



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict

from pydantic import BaseModel, ConfigDict, Field, SkipValidation

from .eventbus.log import LogWriter, TaskLogWriterWrapper
from .lifecycle import Lifecycle

_task_classes = dict()


def task(t: type[Task]):
    if not issubclass(t, Task):
        raise TypeError(f'Must extend `Task` type: {t.__name__}')
    _task_classes[t.__name__] = t
    return t


class Task(ABC, BaseModel, Lifecycle):
    model_config = ConfigDict(extra='forbid', arbitrary_types_allowed=True)
    # Need to skip validation for these, which may contain `multiprocessing.managers` objects. Validation appears to
    # interfere with their functioning...
    shared_dict: SkipValidation[Dict[str, Any]] = Field(default_factory=dict, frozen=True)
    logger: SkipValidation[TaskLogWriterWrapper] = Field(
        default=TaskLogWriterWrapper(LogWriter('Task', None)), frozen=True
    )

    @abstractmethod
    def run(self) -> None:
        pass



from __future__ import annotations

from datetime import datetime, timezone
from enum import Enum
from typing import Iterable, Optional

from pydantic import field_serializer

from . import EventBus, SerializableEvent, serializable_event


class Severity(str, Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'


class SerializableLogEvent(SerializableEvent):
    @classmethod
    def event_group(cls) -> str:
        return 'SerializableLogEvent'


@serializable_event
class LogStartEvent(SerializableLogEvent):
    name: str


@serializable_event
class LogEndEvent(SerializableLogEvent):
    name: str


@serializable_event
class LogWriteEvent(SerializableLogEvent):
    name: str
    severity: Severity
    message: str
    timestamp: datetime

    @field_serializer('timestamp')
    def serialize_timestamp(self, timestamp: datetime, *_):
        return timestamp.isoformat()


@serializable_event
class UnknownLogEvent(SerializableLogEvent):
    content: str


class LogWriter:
    __slots__ = ('bus', 'name')

    def __init__(self, name: str, bus: Optional[EventBus[SerializableLogEvent]]) -> None:
        self.name = name
        self.bus = bus
        if self.bus:
            self.bus.put(LogStartEvent(name=self.name))

    def emit_log_write_event(self, message: str, severity: Severity) -> None:
        if self.bus:
            self.bus.put(LogWriteEvent(
                name=self.name,
                severity=severity,
                message=message,
                timestamp=datetime.now(timezone.utc).astimezone()
            ))

    def close(self) -> None:
        if self.bus:
            self.bus.put(LogEndEvent(name=self.name))


class StdOutLogWriterWrapper:
    __slots__ = ('_base')
    _sev = Severity.INFO

    def __init__(self, log_writer: LogWriter) -> None:
        self._base = log_writer

    def write(self, m: str) -> None:
        if m.strip():
            self._base.emit_log_write_event(m, self._sev)

    def writelines(self, mlist: Iterable[str]) -> None:
        for m in mlist:
            self.write(m)

    def flush(self) -> None:
        # Need to imitate file-like object for redirecting stdout
        pass

    def close(self) -> None:
        # Need to imitate file-like object for redirecting stdout
        pass


class StdErrLogWriterWrapper(StdOutLogWriterWrapper):
    _sev = Severity.ERROR


class TaskLogWriterWrapper:
    __slots__ = ('_base')

    def __init__(self, log_writer: LogWriter) -> None:
        self._base = log_writer

    def debug(self, m: str) -> None:
        self._base.emit_log_write_event(m, Severity.DEBUG)

    def info(self, m: str) -> None:
        self._base.emit_log_write_event(m, Severity.INFO)

    def warning(self, m: str) -> None:
        self._base.emit_log_write_event(m, Severity.WARNING)

    def error(self, m: str) -> None:
        self._base.emit_log_write_event(m, Severity.ERROR)

    def critical(self, m: str) -> None:
        self._base.emit_log_write_event(m, Severity.CRITICAL)



from __future__ import annotations

import asyncio
import inspect
import signal
import sys
import traceback
from contextlib import asynccontextmanager
from multiprocessing import Process
from multiprocessing.managers import DictProxy
from multiprocessing.sharedctypes import Value
from typing import Any, AsyncIterator, Callable, Coroutine, Dict, Optional, TextIO, Type, Union, cast

from .eventbus import EventBus
from .eventbus.execution import ExecutionState, ExecutionStateTransition, SerializableExecutionEvent
from .eventbus.log import (
    LogWriter,
    SerializableLogEvent,
    StdErrLogWriterWrapper,
    StdOutLogWriterWrapper,
    TaskLogWriterWrapper,
)
from .task import Task, _task_classes


async def _default_await_dependencies() -> bool:
    return True


class ProcessResult:
    def __init__(self) -> None:
        self._retcode = Value('i', 0)

    @property
    def is_failed(self) -> bool:
        return bool(self._retcode.value)  # type: ignore

    @is_failed.setter
    def is_failed(self, v: bool) -> None:
        self._retcode.value = v  # type: ignore


def exec_task_lifecycle(
    task_name: str,
    task_class: Type[Task],
    parameters: Optional[Dict[str, Any]],
    log_event_bus: Optional[EventBus[SerializableLogEvent]],
    result: ProcessResult,
    shared_dict: Optional[Union[Dict[str, Any], DictProxy[str, Any]]] = None,
    is_restart: bool = False
):
    base_log_writer = LogWriter(task_name, log_event_bus)
    # Pydantic's BaseModel appears to interfere with the Manager objects when it serializes model values...
    # As a result, any Manager objects should be assigned here directly after being split off into a new process.
    parameters = parameters or dict()
    parameters['logger'] = TaskLogWriterWrapper(base_log_writer)
    parameters['shared_dict'] = cast(Dict[str, Any], shared_dict) if shared_dict is not None else dict()
    task_instance = task_class(**parameters)

    # Bind signal only in new child process
    stdout_log_writer = cast(TextIO, StdOutLogWriterWrapper(base_log_writer))
    stderr_log_writer = cast(TextIO, StdErrLogWriterWrapper(base_log_writer))

    def _exec_lifecycle_stage(stage: Callable[[], None]) -> None:
        try:
            stage()
        except Exception:
            print(traceback.format_exc(), file=stderr_log_writer)
            result.is_failed = True

    signal.signal(signal.SIGTERM, lambda *_: _exec_lifecycle_stage(task_instance.on_abort))
    _sout = sys.stdout
    _serr = sys.stderr

    try:
        sys.stdout = stdout_log_writer
        sys.stderr = stderr_log_writer

        _exec_lifecycle_stage(task_instance.on_create)

        if is_restart:
            _exec_lifecycle_stage(task_instance.on_restart)

        _exec_lifecycle_stage(task_instance.run)

        if result.is_failed:
            _exec_lifecycle_stage(task_instance.on_failure)
        else:
            _exec_lifecycle_stage(task_instance.on_success)
            result.is_failed = False
    except Exception:
        print(traceback.format_exc(), file=sys.stderr)
        result.is_failed = True
    finally:
        _exec_lifecycle_stage(task_instance.on_destroy)
        base_log_writer.close()
        sys.stdout = _sout
        sys.stderr = _serr


class Executor:
    def __init__(
        self,
        name: str,
        task_class: Union[str, Type[Task]],
        log_event_bus: Optional[EventBus[SerializableLogEvent]] = None,
        execution_event_bus: Optional[EventBus[SerializableExecutionEvent]] = None,
        shared_dict: Optional[Union[DictProxy[str, Any], Dict[str, Any]]] = None,
        semaphore: Optional[asyncio.Semaphore] = None,
        max_attempts: int = 1,
        backoff: int = 0,
        await_dependencies: Callable[[], Coroutine[Any, Any, bool]] = _default_await_dependencies,
        is_restart: bool = False,
        parameters: Optional[Dict[str, Any]] = None
    ) -> None:
        self.name = name
        self.log_event_bus = log_event_bus
        self.execution_event_bus = execution_event_bus
        self.shared_dict = shared_dict
        self.max_attempts = max_attempts
        self.semaphore = semaphore
        self.backoff = backoff
        self.task_class = task_class
        self.parameters = parameters
        self.await_dependencies = await_dependencies
        self._state = ExecutionState.INIT
        self.proc: Optional[Process] = None
        self.is_restart = is_restart

    @property
    def state(self) -> ExecutionState:
        return self._state

    @state.setter
    def state(self, val: ExecutionState) -> None:
        event = ExecutionStateTransition(name=self.name, from_state=self._state, to_state=val)
        if self.execution_event_bus is not None:
            self.execution_event_bus.put(event)
        self._state = val

    def get_task_class(self) -> Type[Task]:
        if inspect.isclass(self.task_class) and issubclass(self.task_class, Task):
            return self.task_class
        elif type(self.task_class) == str:
            return _task_classes[self.task_class]
        else:
            raise TypeError('The `task_class` param must be either an extension of `Task` or the string name of one.')

    @asynccontextmanager
    async def acquire_lock(self) -> AsyncIterator[Any]:
        try:
            yield await self.semaphore.acquire() if self.semaphore else None
        finally:
            if self.semaphore:
                self.semaphore.release()

    async def wait(self) -> None:
        await self.event.wait()

    async def start(self) -> None:
        try:
            # Need to wait for main loop to initiate before spawning Event
            self.event = asyncio.Event()

            # In the event of a restart and this task is already complete, return immediately.
            if self.state == ExecutionState.COMPLETED:
                self.event.set()
                return

            # Trigger a state change from INIT -> PENDING
            self.state = ExecutionState.PENDING

            if not await self.await_dependencies():
                self.state = ExecutionState.DEFAULTED
                self.event.set()
                return

            # In the event of skipped task, return immediately.
            if self.state == ExecutionState.SKIP:
                self.event.set()
                return

            attempts = 0
            result = ProcessResult()
            while attempts < self.max_attempts and self.state == ExecutionState.PENDING:
                async with self.acquire_lock():
                    result.is_failed = False
                    self.state = ExecutionState.RUNNING
                    attempts += 1
                    self.proc = Process(
                        target=exec_task_lifecycle,
                        args=(
                            self.name,
                            self.get_task_class(),
                            self.parameters,
                            self.log_event_bus,
                            result,
                            self.shared_dict,
                            self.is_restart
                        ),
                        daemon=False
                    )
                    self.proc.start()
                    loop = asyncio.get_running_loop()
                    await loop.run_in_executor(None, self.proc.join)

                # Restart check
                if result.is_failed and (attempts < self.max_attempts):
                    self.state = ExecutionState.PENDING
                    await asyncio.sleep(self.backoff)

            if result.is_failed:
                self.state = ExecutionState.FAILED
            else:
                self.state = ExecutionState.COMPLETED
        except asyncio.CancelledError:
            self.terminate()
            self.state = ExecutionState.ABORTED
        finally:
            self.event.set()

    def terminate(self) -> None:
        if self.proc is not None:
            # Send SIGTERM to child process
            self.proc.terminate()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from ..eventbus.log import SerializableLogEvent
from ..lifecycle import AsyncLifecycle

_logger_classes = dict()


def logger(t: type[Logger]) -> Any:
    if not issubclass(t, Logger):
        raise TypeError(f'Must extend `Logger` type: {t.__name__}')
    _logger_classes[t.__name__] = t
    return t


class Logger(ABC, AsyncLifecycle, BaseModel):
    model_config = ConfigDict(extra='forbid')

    @abstractmethod
    async def update(self, m: SerializableLogEvent) -> None:
        pass



import json

import requests

from ..extension import extension
from .notification import Notification


@extension
class SlackWebhookNotification(Notification):
    webhook: str

    async def send_notification(self, title: str, msg: str) -> None:
        requests.post(
            self.webhook,
            data=json.dumps({'text': title, 'attachments': [{'text': msg}]}),
            headers={'Content-Type': 'application/json'},
        )



class NoTasksLoadedError(Exception):
    pass


class ExtensionsDirectoryNotFoundError(Exception):
    pass


class NotAPackageError(Exception):
    pass


class CheckpointInvalidError(Exception):
    pass



from abc import abstractmethod
from datetime import datetime

from ...executor import SerializableExecutionEvent
from ..extension import Extension


class Notification(Extension):
    notify_on_create: bool = True
    notify_on_success: bool = True
    notify_on_failure: bool = True
    notify_on_abort: bool = True

    async def update(self, _: SerializableExecutionEvent) -> None:
        # We don't want notifications to be spammed...
        pass

    @abstractmethod
    async def send_notification(self, title: str, msg: str) -> None:
        pass

    async def on_create(self) -> None:
        if self.notify_on_create:
            await self.send_notification(
                "Flowmancer Job Notification: STARTING",
                f"Job initiated at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            )

    async def on_success(self) -> None:
        if self.notify_on_success:
            await self.send_notification(
                "Flowmancer Job Notification: SUCCESS",
                f"Job completed successfully at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}",
            )

    async def on_failure(self) -> None:
        if self.notify_on_failure:
            await self.send_notification(
                "Flowmancer Job Notification: FAILURE", f"Job failed at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            )

    async def on_abort(self) -> None:
        if self.notify_on_abort:
            await self.send_notification(
                "Flowmancer Job Notification: ABORTED", f"Job aborted at {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            )



import os
import re
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Pattern, Union

import yaml

from . import JobDefinition, LoadParams, SerializableJobDefinition, job_definition


def _merge(a: Any, b: Any) -> None:
    for k in b.keys():
        if k in a and isinstance(a[k], dict) and isinstance(b[k], dict):
            _merge(a[k], b[k])
        else:
            a[k] = b[k]


def _build_path_constructor(var_matcher: Pattern[str], val_dict: Dict[str, Any]) -> Callable:
    def _path_constructor(_: yaml.SafeLoader, node: yaml.Node) -> str:
        def replace_fn(match):
            parts = f"{match.group(1)}:".split(":")
            return val_dict.get(parts[0], parts[1])
        return var_matcher.sub(replace_fn, node.value)
    return _path_constructor


@job_definition('yaml')
class YAMLJobDefinition(SerializableJobDefinition):
    def load(
        self, filename: Union[Path, str], params: LoadParams = LoadParams(), vars: Optional[Dict[str, str]] = None
    ) -> JobDefinition:
        # Add constructor for built-in vars
        sys_tag = re.compile(r'[^$]*\$SYS{([^}^{]+)}.*')
        sys_var = re.compile(r'\$SYS{([^}^{]+)}')
        yaml.add_implicit_resolver("!sysvar", sys_tag, None, yaml.SafeLoader)
        yaml.add_constructor("!sysvar", _build_path_constructor(sys_var, dict(params)), yaml.SafeLoader)

        # Add constructor for env vars
        env_tag = re.compile(r'[^$]*\$ENV{([^}^{]+)}.*')
        env_var = re.compile(r'\$ENV{([^}^{]+)}')
        yaml.add_implicit_resolver("!envvar", env_tag, None, yaml.SafeLoader)
        yaml.add_constructor("!envvar", _build_path_constructor(env_var, dict(os.environ)), yaml.SafeLoader)

        # Add constructor for input vars
        input_tag = re.compile(r'[^$]*\$VAR{([^}^{]+)}.*')
        input_var = re.compile(r'\$VAR{([^}^{]+)}')
        yaml.add_implicit_resolver("!inputvar", input_tag, None, yaml.SafeLoader)
        yaml.add_constructor("!inputvar", _build_path_constructor(input_var, vars or dict()), yaml.SafeLoader)

        def process_includes(jdef, merged, seen):
            for p in jdef.get('include', []):
                if not p.startswith('/'):
                    p = os.path.abspath(os.path.join(params.APP_ROOT_DIR, p))
                if p in seen:
                    raise RuntimeError(f'JobDef YAML file has already been processed once: {p}')
                seen.add(p)
                with open(p, 'r') as f:
                    cur = yaml.safe_load(f.read())
                    process_includes(cur, merged, seen)
            _merge(merged, jdef)

        merged: Dict[str, Any] = dict()
        with open(filename, 'r') as f:
            process_includes(yaml.safe_load(f.read()), merged, set())
        return JobDefinition(**merged)

    def dump(self, jdef: JobDefinition, filename: Union[Path, str]) -> None:
        with open(filename, 'r') as f:
            f.write(yaml.dump(jdef.model_dump()))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flowmancer',
 'flowmancer.checkpointer',
 'flowmancer.eventbus',
 'flowmancer.extensions',
 'flowmancer.extensions.notifications',
 'flowmancer.jobdefinition',
 'flowmancer.loggers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0.1,<7.0.0',
 'pydantic>=2.6.3,<3.0.0',
 'requests>=2.31.0,<3.0.0',
 'rich>=13.7.1,<14.0.0',
 'types-pyyaml>=6.0.12.12,<7.0.0.0',
 'urllib3>=2.2.1,<3.0.0']

setup_kwargs = {
    'name': 'flowmancer',
    'version': '0.11.1',
    'description': 'The Python Thing-Doer',
    'long_description': '# Flowmancer\n\n[![pypi-version](https://img.shields.io/pypi/v/flowmancer?style=flat-square)](https://pypi.org/project/flowmancer)\n[![python-version](https://img.shields.io/badge/dynamic/json?query=info.requires_python&label=python&url=https%3A%2F%2Fpypi.org%2Fpypi%2Fflowmancer%2Fjson&style=flat-square)](https://pypi.org/project/flowmancer)\n[![license](https://img.shields.io/github/license/natsunlee/flowmancer?style=flat-square)](LICENSE)\n[![circle-ci](https://img.shields.io/circleci/build/github/natsunlee/flowmancer?style=flat-square)](https://app.circleci.com/pipelines/github/natsunlee/flowmancer)\n[![coveralls](https://img.shields.io/coveralls/github/natsunlee/flowmancer?style=flat-square)](https://coveralls.io/github/natsunlee/flowmancer?branch=main)\n[![pypi-downloads](https://img.shields.io/pypi/dm/flowmancer?style=flat-square)](https://pypistats.org/packages/flowmancer)\n[![Ko-Fi](https://img.shields.io/badge/Support%20Me%20On%20Ko--fi-F16061?style=flat-square&logo=ko-fi&logoColor=white)](https://ko-fi.com/natsunlee)\n\nFlowmancer aims to help you do *things* in a sequential or parallel manner. It enables you to write tasks in Python, describe their order, then execute them with as little effort as possible.\n\nBut why do I need this? Couldn\'t I just write my own Python code to do *stuff*?\n\nYou certainly could!\n\nThough Flowmancer provides gives you a head-start to building your custom processes with optional add-ons for logging, checkpoint/restarts in the event of failures, or even custom task observers to do...things while your things do things!\n\n## Installation\nSimply install the `flowmancer` package with:\n```bash\npip3 install flowmancer\n```\n\nNOTE: `flowmancer` supports only Python 3.7 and higher.\n\n## Basic Usage\nLet\'s assume you have a new project with a basic structure like so:\n```\nmy_project\n├─ job.yaml\n├─ main.py\n└─ tasks/\n   ├─ __init__.py\n   └─ mytasks.py\n```\n\nTo use `flowmancer`, you\'ll need to provide a few things:\n* `Task` implementations (`mytasks.py`)\n* A job YAML file (`job.yaml`)\n* Your main/driver code (`main.py`)\n\n### Tasks\nBy default, Flowmancer recursively searches in the `./tasks` directory (relative to where `Flowmancer()` is initialized - in this case, `main.py`) for `Task` implementations decorated with `@task`. See the Advanced Usage section for details on how to add other directories or packages that contain `Task` implementations.\n\nA `flowmancer` task is simply a class that extends the `Task` abstract class, which, at minimum requires that the `run` method be implemented:\n```python\nimport time\nfrom flowmancer.task import Task, task\n\n@task\nclass WaitAndSucceed(Task):\n    # All variables should be given type hints and optional vars should be given default values.\n    my_required_string_var: str\n    my_optional_int_var: int = 5\n\n    def run(self):\n        # Store string input var in the shared dictionary accessible by other tasks.\n        self.shared_dict["my_var"] = f"Hello from: {self.my_required_string_var}!"\n\n        # Sleep for seconds defined by input var (using default of 5).\n        print(f"Starting up and sleeping for {self.my_optional_int_var} seconds!")\n        time.sleep(self.my_optional_int_var)\n        print("Done!")\n\n@task\nclass ImmediatelySucceed(Task):\n    def run(self):\n        # Print statements will automatically be sent to configured loggers.\n        print("Success!")\n\n@task\nclass FailImmediately(Task):\n    def run(self):\n        print(f"Printing `my_var` value: {self.shared_dict[\'my_var\']}")\n        # Raise errors to cause tasks to fail and additionally block dependent tasks, if any.\n        raise RuntimeError("Let this be caught by Flowmancer")\n```\n\nAny `print()` or exceptions will write log messages to any configured loggers (zero or more loggers may be defined).\n\n### Job Definition YAML File\nThis file describes what code to run, in what order, as well as additional add-ons to supplement the job during execution:\n```yaml\nversion: 0.1\n\n# This entire config block is currently optional, however, it is recommended to at least provide a unique name for each\n# Job Definition YAML file, as this name is used for checkpointing jobs in the event of failures.\nconfig:\n  name: \'my-flowmancer-job\'\n\ntasks:\n  # No dependency - run right away\n  # Add `parameters` key-value pairs for any required and optional task variables.\n  succeed-task-a:\n    task: WaitAndSucceed\n    parameters:\n      my_required_string_var: "My First Task!"\n\n  # No dependency - run right away\n  succeed-task-b:\n    task: ImmediatelySucceed\n\n  # Only run if prior 2 tasks complete successfully\n  final-fail-task:\n    task: FailImmediately\n    max_attempts: 3  # Retry uup to 2 times upon failure (1 initial exec + 2 retries = 3 attempts)\n    dependencies:\n      - succeed-task-a\n      - succeed-task-b\n```\n\n### Driver\nThe driver is super simple and simply requires running an instance of `Flowmancer`\n```python\n# main.py\nimport sys\nfrom flowmancer import Flowmancer\n\nif __name__ == \'__main__\':\n    # The `start()` method will return a non-zero integer on failure, typically equal to the number of failed tasks.\n    ret = Flowmancer().start()\n\n    # Exceptions from tasks will be captured and logged, rather than being raised up to this level. To cause this\n    # driver program to fail, either explicitly raise your own error OR call sys.exit.\n    if ret:\n      raise RuntimeError(\'Flowmancer job has failed!\')\n\n    # Alternatively, instead of crashing w/ an exception, simply exit with a non-zero value.\n    # sys.exit(ret)\n```\n\n### Executing the Job\n```bash\npython3 main.py -j ./path/to/job.yaml\n```\n\nTo run from point-of-failure (if any):\n```bash\npython3 main.py -j ./path/to/job.yaml -r\n```\nIf no prior failure is detected, the job will start as if no `-r` flag were given.\n\nNote that the job definition must still be provided with the `-r` flag.\n\n## Advanced Usage\n\n### Optional Configurations\nIn the `config` block of the Job Definition, the following optional parameters may be given:\n|Parameter|Type|Default Value|Description|\n|---|---|---|---|\n|name|str|\'flowmancer\'|Name/identifier for Job Definition. Used for saving checkpoints used for job restarts in the event of a failure.|\n|max_concurrency|int|0|Maximum number tasks that can run in parallel. If 0 or less, then there is no limit.|\n|extension_directories|List[str]|[]|List of paths, either absolute or relative to driver `.py` file, that contain any `@task`, `@logger`, or `@extension` decorated classes to make accessible to Flowmancer. The `./task`, `./extensions`, and `./loggers` directories are ALWAYS checked by default.|\n|extension_packages|List[str]|[]|List of installed Python packages that contain `@task`, `@logger`, or `@extension` decorated classes to make accessible to Flowmancer.|\n|synchro_interval_seconds|float|0.25|Core execution loop interval for waking and checking status of tasks and whether loggers/extensions/checkpointer should trigger.|\n|loggers_interval_seconds|float|0.25|Interval in seconds to wait before emitting log messages to configured `Logger` instances.|\n|extensions_interval_seconds|float|0.25|Interval in seconds to wait before emitting state change information to configured `Extension` instances.|\n|checkpointer_interval_seconds|float|10.0|Interval in seconds to wait before writing checkpoint information to the configured `Checkpointer`.|\n\nFor example:\n```yaml\nconfig:\n  name: \'most-important-job\'\n  max_concurrency: 20\n  extension_directories:\n    - ./client_implementations\n    - /opt/flowmancer/tasks\n  extension_packages:\n    - internal_flowmancer_package\n```\n\n### Include YAML Files\nAn optional `include` block may be defined in the Job Definition in order to merge multiple Job Definition YAML files.\nYAML files are provided in a list and processed in the order given, with the containing YAML being processed last.\n\nFor example:\n```yaml\n# <app_root_dir>/jobdefs/template.yaml\nconfig:\n  name: generic-template\n\ntasks:\n  do-something:\n    task: DoSomething\n    parameters:\n      some_required_param: I am a required string parameter\n```\n\n```yaml\n# <app_root_dir>/jobdefs/cleanup_addon.yaml\ninclude:\n  - ./jobdefs/template.yaml\n\ntasks:\n  cleanup:\n    task: Cleanup\n    dependencies:\n      - do-something\n```\n\n```yaml\n# <app_root_dir>/jobdefs/complete.yaml\nconfig:\n  name: complete-job\n\ninclude:\n  - ./jobdefs/cleanup_addon.yaml\n\ntasks:\n  do-something:\n    task: Do Something\n    parameters:\n      added_optional_param: 99\n```\n\nLoading the `complete.yaml` job definition will result in a YAML equivalent to:\n```yaml\nconfig:\n  name: complete-job\n\ntasks:\n  do-something:\n    task: Do Something\n    parameters:\n      some_required_param: I am a required string parameter\n      added_optional_param: 99\n```\n\n> :warning: Array values are **NOT** merged like dictionaries are. Any array values (and therfore any nested structures) within them will be replaced if modified in a later YAML.\n\nAdditionally, the above example could have all `include` values in the `complete.yaml` file and the `include` block removed from `cleanup_addon.yaml`:\n```yaml\n# <app_root_dir>/jobdefs/complete.yaml\nconfig:\n  name: complete-job\n\n# As with most paths in Job Definition, paths to `include` YAML files are relative to `.py` file where the `.start()`\n# method for Flowmancer is invoked.\ninclude:\n  - ./jobdefs/template.yaml\n  - ./jobdefs/cleanup_addon.yaml\n\ntasks:\n  do-something:\n    task: Do Something\n    parameters:\n      added_optional_param: 99\n```\n\nThe `include` values are processed in order and results in the same outcome as the original example.\n\n### Changing Default File Logger Directory\nThe Job Definition accepts an optional `loggers` section, which if left empty will default to using a `FileLogger` with default settings.\nTo utilize the default `FileLogger`, but with a different configuration, explicitly provide the `loggers` block:\n```yaml\nloggers:\n  my-file-logger:\n    logger: FileLogger\n    parameters:\n      # NOTE: this path is relative to the `.py` file where `Flowmancer().start()` is invoked.\n      base_log_dir: ./my_custom_log_dir  # ./logs is the default, if omitted.\n      retention_days: 3  # 10 is the default, if omitted.\n```\n\n### Complex Parameters\nWhile this is mostly used for `Task` implementations, the details outlined here apply for any built-in and custom `Extension` and `Logger` implementations.\n\nFlowmancer makes heavy use of [Pydantic](https://docs.pydantic.dev/latest/) to validate parameters and ensure that values loaded from the Job Definition are of the appropriate type.\n\nThis means that a `Task` can have complex types (including custom models) like:\n```python\nfrom enum import Enum\nfrom flowmancer.task import Task, task\nfrom pydantic import BaseModel\nfrom typing import Dict, List\n\nclass Protocol(str, Enum):\n    HTTP: \'HTTP\'\n    HTTPS: \'HTTPS\'\n\nclass APIDetails(BaseModel):\n    protocol: Protocol = Protocol.HTTPS\n    base_url: str\n    endpoint: str\n\n@task\nclass DownloadDataFromRestApi(Task):\n    api_details: APIDetails\n    target_dir: str\n    target_filename: str = \'data.json\'\n\n    def run(self) -> None:\n        url = f\'{self.api_details.protocol}://{self.api_details.base_url}/{self.api_details.endpoint}\'\n        # Continued implementation...\n```\n\nAnd the Job Definition snippet for this task might be:\n```yaml\ntasks:\n  download-file-one:\n    task: DownloadDataFromRestApi\n    parameters:\n      api_details:\n        # We leave out `protocol` because we want to just use the default `HTTPS` value.\n        base_url: www.some_data_api.com\n        endpoint: /v1/data/weather/today\n      target_dir: /data/todays_weather\n      # Override the default `target_filename` value given in the class implementation.\n      target_filename: weather.json\n```\n\n### Task Lifecycle Methods\nIn addition to the required `run` method, an implementation of `Task` may optionally include the following methods:\n|Method|Required|Order|Description|\n|---|---|---|---|\n|on_create|No|1|First method executed when a task is released for execution. Note that a task is not considered "created" until it enters the `RUNNING` state.|\n|on_restart|No|2|Executed only if a task is running from the result of a recovery from `FAILED` state. If a task was failed in `DEFAULTED` state, this method will not be executed.|\n|run|Yes|3|Always required and always executed once task is in `RUNNING` state, unless prior lifecycle methods have failed.|\n|on_success|No|4|Executed only if `run` method ends in success.|\n|on_failure|No|5|Executed only if `run` method ends in failure/exception.|\n|on_destroy|No|6|Always executed after all other lifecycle methods.|\n|on_abort|No|-|Executed when `SIGINT` signal is sent to tasks/Flowmancer.|\n\nJust as with `run`, all lifecycle methods have access to `self.shared_dict` and any parameters.\n\n### Custom Loggers\nCustom implementations of the `Logger` may be provided to Flowmancer to either replace OR write to in addition to the default `FileLogger`.\n\nA custom implementation must extend the `Logger` class, be decorated with the `logger` decorator, and implement the async `update` method at minimum:\n```python\n@logger\nimport json\nimport requests\nfrom flowmancer.loggers.logger import Logger, logger\nfrom flowmancer.eventbus.log import LogEndEvent, LogStartEvent, LogWriteEvent, SerializableLogEvent\n\nclass SlackMessageLogger(Logger):\n    webhook: str\n\n    def _post_to_slack(self, msg: str) -> None:\n        requests.post(\n            self.webhook,\n            data=json.dumps({\'text\': title, \'attachments\': [{\'text\': msg}]}),\n            headers={\'Content-Type\': \'application/json\'},\n        )\n\n    async def update(self, evt: SerializableLogEvent) -> None:\n        # The `LogStartEvent` and `LogEndEvent` events only have a `name` property.\n        if isinstance(evt, LogStartEvent):\n            self._post_to_slack(f\'[{evt.name}] START: Logging is beginning\')\n        elif isinstance(evt, LogEndEvent):\n            self._post_to_slack(f\'[{evt.name}] END: Logging is ending\')\n        # The `LogWriteEvent` additionally has `severity` and `message` properties.\n        elif isinstance(evt, LogWriteEvent):\n            self._post_to_slack(f\'[{evt.name}] {evt.severity.value}: {evt.message}\')\n```\n\nThe `Logger` implementation may also have the following optional `async` lifecycle methods:\n* `on_create`\n* `on_restart`\n* `on_success`\n* `on_failure`\n* `on_destroy`\n* `on_abort`\n\nTo incorporate your custom `Logger` into Flowmancer, ensure that it exists in a module either in `./loggers` or in a module listed in `config.extension_directories` in the Job Definition.\n\nThis allows it to be provided in the `loggers` section of the Job Definition.\n> :warning: Providing the `loggers` section will remove the default logger (`FileLogger`) from your job\'s configuration.\n> If you want to add your custom logger alongside the default logger, the `FileLogger` must explicitly be configured.\n\n```yaml\nloggers:\n  # Load the default logger with default parameters\n  default-logger:\n    logger: FileLogger\n\n  # Custom logger implementation\n  slack-logger:\n    logger: SlackMessageLogger\n    parameters:\n      webhook: https://some.webhook.url\n```\n\n### Custom Extensions\nComing soon.\n\n### Custom Checkpointers\nCustom implementations of the `Checkpointer` may be provided to Flowmancer to replace the default `FileCheckpointer`.\n> :warning: Unlike loggers and extensions, only one checkpointer can be configured per Job Definition.\n\nA custom implementation must extend the `Checkpointer` class, be decorated with the `checkpointer` decorator, and implement the async `write_checkpoint`, `read_checkpoint`, and `clear_checkpoints` methods at minimum. It may also optinoally implement async lifecycle methods, similar to [Custom Loggers](#custom-loggers):\n```python\nfrom .checkpointer import CheckpointContents, Checkpointer, NoCheckpointAvailableError, checkpointer\n\n@checkpointer\nclass DatabaseCheckpointer(Checkpointer):\n    host: str\n    port: int\n    username: str\n    password: str\n\n    def write_checkpoint(self, name: str, content: CheckpointContents) -> None:\n        # Store checkpoint state - must be able to store contents of\n        # `CheckpointContents` in a way that it can be reconstructed later.\n\n    def read_checkpoint(self, name: str) -> CheckpointContents:\n        # Recall checkpoint state - reconstruct and return `CheckpointContents`\n        # if exists for `name`. Otherwise raise `NoCheckpointAvailableError`\n        # to indicate no valid checkpoint exists to restart from.\n\n    def clear_checkpoint(self, name: str) -> None:\n        # Remove checkpoint state for `name`.\n```\n\nTo incorporate your custom `Checkpointer` into Flowmancer, ensure that it exists in a module either in `./extensions` or in a module listed in `config.extension_directories` in the Job Definition.\n\nThis allows it to be provided in the `checkpointer` section of the Job Definition:\n```yaml\ncheckpointer:\n  checkpointer: DatabaseCheckpointer\n  parameters:\n    host: something\n    port: 9999\n    username: user\n    password: 1234\n```\n',
    'author': 'Nathan Lee',
    'author_email': 'lee.nathan.sh@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from enum import Enum
from typing import Dict, Set, Union

from . import SerializableEvent, serializable_event


class SerializableExecutionEvent(SerializableEvent):
    @classmethod
    def event_group(cls) -> str:
        return 'SerializableExecutionEvent'


class ExecutionState(str, Enum):
    FAILED = 'F'
    PENDING = 'P'
    RUNNING = 'R'
    DEFAULTED = 'D'
    COMPLETED = 'C'
    ABORTED = 'A'
    SKIP = 'S'
    INIT = '_'


class ExecutionStateMap:
    def __init__(self) -> None:
        self.data: Dict[ExecutionState, Set[str]] = dict()

    def __getitem__(self, k: Union[str, ExecutionState]) -> Set[str]:
        es = ExecutionState(k)
        if es not in self.data:
            self.data[es] = set()
        return self.data[ExecutionState(k)]

    def __setitem__(self, k: Union[str, ExecutionState], v: Set[str]) -> None:
        es = ExecutionState(k)
        self.data[es] = v

    def __str__(self):
        return str(self.data)

    def items(self):
        return self.data.items()

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    @classmethod
    def from_simple_dict(cls, data: Dict[str, Set[str]]) -> ExecutionStateMap:
        m = ExecutionStateMap()
        for k, v in data.items():
            m[ExecutionState(k)] = v
        return m

    def to_simple_dict(self) -> Dict[str, Set[str]]:
        return {k.value: v for k, v in self.data.items()}


@serializable_event
class ExecutionStateTransition(SerializableExecutionEvent):
    name: str
    from_state: ExecutionState
    to_state: ExecutionState



from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type, Union

from pydantic import BaseModel, ConfigDict

_job_definition_classes = dict()


def job_definition(key: str) -> Callable:
    def inner(t: type[SerializableJobDefinition]) -> Type[SerializableJobDefinition]:
        if not issubclass(t, SerializableJobDefinition):
            raise TypeError(f'Must extend `SerializableJobDefinition` type: {t.__name__}')
        _job_definition_classes[key] = t
        return t
    return inner


class JobDefinitionComponent(BaseModel):
    model_config = ConfigDict(extra='forbid', use_enum_values=True)


class LoggerDefinition(JobDefinitionComponent):
    logger: str
    parameters: Dict[str, Any] = dict()


class TaskDefinition(JobDefinitionComponent):
    task: str
    dependencies: List[str] = []
    max_attempts: int = 1
    backoff: int = 0
    parameters: Dict[str, Any] = dict()


class ExtensionDefinition(JobDefinitionComponent):
    extension: str
    parameters: Dict[str, Any] = dict()


class ConfigurationDefinition(JobDefinitionComponent):
    name: str = 'flowmancer'
    max_concurrency: int = 0
    extension_directories: List[str] = []
    extension_packages: List[str] = []
    synchro_interval_seconds: float = 0.25
    loggers_interval_seconds: float = 0.25
    extensions_interval_seconds: float = 0.25
    checkpointer_interval_seconds: float = 10.0


class CheckpointerDefinition(JobDefinitionComponent):
    checkpointer: str
    parameters: Dict[str, Any] = dict()


class JobDefinition(JobDefinitionComponent):
    version: float = 0.1
    include: List[Path] = []
    config: ConfigurationDefinition = ConfigurationDefinition()
    tasks: Dict[str, TaskDefinition]
    loggers: Dict[str, LoggerDefinition] = {'file-logger': LoggerDefinition(logger='FileLogger')}
    extensions: Dict[str, ExtensionDefinition] = {'progress-bar': ExtensionDefinition(extension='RichProgressBar')}
    checkpointer: CheckpointerDefinition = CheckpointerDefinition(checkpointer='FileCheckpointer')


class LoadParams(BaseModel):
    model_config = ConfigDict(extra='forbid')
    APP_ROOT_DIR: str = '.'


class SerializableJobDefinition(ABC):
    @abstractmethod
    def load(
        self, filename: Union[Path, str], params: LoadParams = LoadParams(), vars: Optional[Dict[str, str]] = None
    ) -> JobDefinition:
        pass

    @abstractmethod
    def dump(self, jdef: JobDefinition, filename: Union[Path, str]) -> None:
        pass



import smtplib
from email.message import EmailMessage

from ..extension import extension
from .notification import Notification


@extension
class EmailNotification(Notification):
    recipient: str
    smtp_host: str
    smtp_port: int
    sender_user: str
    sender_host: str

    async def send_notification(self, title: str, msg: str) -> None:
        em = EmailMessage()
        em['From'] = f'{self.sender_user}@{self.sender_host}'
        em['Subject'] = title
        em['To'] = self.recipient
        em.set_content(msg)
        with smtplib.SMTP(self.smtp_host, self.smtp_port) as s:
            s.send_message(em)



import os
import pickle
from pathlib import Path

from .checkpointer import CheckpointContents, Checkpointer, NoCheckpointAvailableError, checkpointer


@checkpointer
class FileCheckpointer(Checkpointer):
    checkpoint_dir: str = './.flowmancer'

    async def write_checkpoint(self, name: str, content: CheckpointContents) -> None:
        cdir = Path(self.checkpoint_dir)
        if not os.path.exists(cdir):
            os.makedirs(cdir, exist_ok=True)
        tmp = cdir / (name + '.tmp')
        perm = Path(self.checkpoint_dir) / name
        with open(tmp, 'wb') as f:
            pickle.dump(content, f)
        if os.path.isfile(perm):
            os.unlink(perm)
        os.rename(tmp, perm)

    async def read_checkpoint(self, name: str) -> CheckpointContents:
        checkpoint_file = Path(self.checkpoint_dir) / name
        if not checkpoint_file.exists():
            raise NoCheckpointAvailableError(f'Checkpoint file does not exist: {checkpoint_file}')
        return pickle.load(open(checkpoint_file, 'rb'))

    async def clear_checkpoint(self, name: str) -> None:
        cfile = Path(self.checkpoint_dir) / name
        if os.path.isfile(cfile):
            os.unlink(cfile)
